# -*- coding: utf-8 -*-

# magnifier.py
# Version: python v.3.x

# Author: Vincent Vande Vyvre <vincent.vandevyvre@swing.be>
# Licence: GPL3

import os
import sys

from PyQt4 import QtCore, QtGui


class Magnifier(QtGui.QGraphicsPixmapItem):
    def __init__(self, main, parent=None, graphic=None, scene=None, img=None, zoom=None):
        super(Magnifier, self).__init__(parent, scene)
        self.main = main
        d = dict(shape = "circle",
                    size = 160,
                    factor = 1.0,
                    image = img,
                    graphic = graphic,
                    scene = scene,
                    cur_pos = None,
                    zoom = zoom,
                    background_color = (255, 255, 255, 255),
                    old_X = None,
                    old_Y = None)
        for key, value in d.items():
            setattr(self, key, value)

        self.setFlags(QtGui.QGraphicsItem.ItemIsMovable |
                        QtGui.QGraphicsItem.ItemIsFocusable)

        self.setAcceptHoverEvents(True)

        self.build_settings()
        self._moved = Moved()
        self._optionChanged = OptionChanged()


    # Create signals
    # Signals are class defined at the end of this script
    # http://docs.python.org/howto/descriptor.html

    def positionChanged():
        def fget(self):
            return self._moved.positionChanged
        return locals()
    positionChanged = property(**positionChanged())

    def settingsChanged():
        def fget(self):
            return self._optionChanged.settingsChanged
        return locals()
    settingsChanged = property(**settingsChanged())

    def set_size(self, size):
        """Set the size of the magnifying glass.

        The size represent the side of the shape 'square' or the diameter of
        the shape 'circle'

        Keyword arguments:
        size -- integer
        """
        if not isinstance(size, int):
            print("Magnifier.set_size arg not int, got %s" % size)
            return
        self.size = size
        if self.cur_pos:
            self.__redraw()

    def set_shape(self, shape):
        """Set the shape of the magnifying glass.

        Keyword arguments:
        shape -- str() "square" or "circle"
        """
        if shape not in ["square", "circle"]:
            print("Magnifier.set_shape arg must be 'square' or 'circle',"\
                                " got %s" % shape)
            return
        self.shape = shape
        if self.cur_pos:
            self.__redraw()

    def set_factor(self, fact):
        """Set the magnification factor.

        Keyword arguments:
        fact -- float
        """
        if not isinstance(fact, float) and not fact == 1:
            print("Magnifier.set_factor arg must be float or 1, got %s" % fact)
            return
        self.factor = float(fact)
        if self.cur_pos:
            p = self.pos()
            self.setup()
            p2 = self.pos()
            self.setPos(p)
            x, y = p.x() - p2.x(), p.y() - p2.y()
            self.__update(x, y)

    def set_background_color(self, color):
        """Set the background color of the magnifying glass.

        Set the background color of the magnifying glass with the background
        color of the QGraphicsView.
        When the magnifying glass leave the image we fill the texture with
        this color.
        This only needed with the shape 'circle'.

        Keyword arguments:
        color -- tuple or list, if rgb is given, alpha = 255 is append.
        """
        try:
            r, v, b = color[0:3]
            c = QtGui.QColor()
            c.setRgb(r, v, b)
        except:
            print("Magnifier.set_background_color value error, got %s" % color)
            return
        if len(color) == 3:
            # change rgb to rgba
            color = [c for c in color[:]]
            color.append(255)
        self.background_color = color
        self._set_brush()

    def _set_brush(self):
        # Qcolor don't like tuple
        r, v, b, a = self.background_color
        color = QtGui.QColor()
        color.setRgb(r, v, b, a)
        self.brush = QtGui.QBrush(color, QtCore.Qt.SolidPattern)

    def setup(self):
        """setup the magnifier.

        """
        if not self.graphic:
            print("\nWarning: Magnifier.setup() no graphic given. Abort")
            return
        if not self.scene:
            print("\nWarning: Magnifier.setup() no scene given. Abort")
            return
        if not self.image:
            print("\nWarning: Magnifier.setup() no image given. Abort")
            return
        if not self.zoom:
            print("\nWarning: Magnifier.setup() no zoom given. Abort")
            return

        self.cur_pos = self.__get_center()
        self.setPos(self.cur_pos[2], self.cur_pos[3])
        self._set_brush()
        self.multi = self.factor * (1 / self.zoom)
        self.__create_magnifier()
        if not self.old_X:
            self.recenter_magnifying_glass()
        self.old_X, self.old_Y = self.pos().x(), self.pos().y()


    def __create_magnifier(self):
        """Call functions to build the magnifying glass.

        """
        self.__set_pixmap()
        self.orig = self.__get_orig_center()
        if self.shape == "square":
            self.__set_square()
        else:
            self.__set_circle()

    def __set_pixmap(self):
        """Resize the image with magnification factor.

        """
        w, h = self.image.width(), self.image.height()
        self.base = self.image.scaled(w * self.factor, h * self.factor,
                                    QtCore.Qt.KeepAspectRatio,
                                    QtCore.Qt.SmoothTransformation)

    def __set_pix(self):
        """Create the pixmap support for the texture.

        Returns:
        QPixmap square filled transparent
        """
        color = QtGui.QColor()
        color.setRgb(0, 0, 0, 0)
        pix = QtGui.QPixmap(QtCore.QSize(self.size + 4, self.size + 4))
        pix.fill(color)
        return pix

    def __set_square(self):
        """Build a squared magnifying glass.

        """
        self.pix = self.__set_pix()
        rect = QtCore.QRect(0, 0, self.size + 4, self.size + 4)
        painter = QtGui.QPainter()
        painter.begin(self.pix)
        painter.setPen(QtGui.QPen(QtCore.Qt.blue, 2,
                                  QtCore.Qt.SolidLine, QtCore.Qt.RoundCap,
                                  QtCore.Qt.MiterJoin))
        painter.drawRect(rect)
        painter.end()
        self.__fill_square()

    def __fill_square(self):
        """Draw the texture into the shape 'square'.

        """
        painter = QtGui.QPainter()
        painter.begin(self.pix)
        rect = QtCore.QRect(2, 2, self.size, self.size)
        painter.eraseRect(rect)
        irect = QtCore.QRect(self.orig[2], self.orig[3], self.size, self.size)
        painter.drawImage(rect, self.base, irect)
        painter.end()
        self.setPixmap(self.pix)

    def __set_circle(self):
        """Build a circular magnifying glass.

        """
        self.pix = self.__set_pix()
        rect = QtCore.QRect(1, 1, self.size + 3, self.size + 3)
        painter = QtGui.QPainter()
        painter.begin(self.pix)
        painter.setRenderHints(QtGui.QPainter.Antialiasing, True)
        painter.setPen(QtGui.QPen(QtCore.Qt.blue, 1,
                                  QtCore.Qt.SolidLine, QtCore.Qt.RoundCap,
                                  QtCore.Qt.MiterJoin))
        painter.drawEllipse(rect)
        rect = QtCore.QRectF(2, 2, self.size, self.size)
        self.path = QtGui.QPainterPath()
        self.path.addEllipse(rect)
        painter.drawPath(self.path)
        painter.end()
        self.__fill_circle()

    def __fill_circle(self):
        """Draw the texture into the shape 'circle'.

        """
        painter = QtGui.QPainter()
        brush = QtGui.QBrush()
        brush.setTexture(self.__set_texture())
        painter.begin(self.pix)
        painter.fillPath(self.path, brush)
        painter.end()
        self.setPixmap(self.pix)

    def __update(self, x, y):
        """Update the original coordinates when the magnifying glass is moved.

        """
        self.orig[2] += x * self.multi
        self.orig[3] += y * self.multi
        if self.shape == "square":
            self.__fill_square()
        else:
            self.__fill_circle()

    def move_by(self, x, y):
        """Move the magnifying glass.

        Keyword arguments:
        x -- number of pixels on x axis.
        y -- number of pixels on y axis.
        """
        if not isinstance(x, int) or not isinstance(y, int):
            print("Magnifier.move_by values not integer, got {0}, {1}"\
                                                .format(x, y))
            return
        self.moveBy(x, y)
        self.__update(x, y)

    def __redraw(self):
        """Redraw magnifying glass when parameters are changed.

        """
        self.prepareGeometryChange()
        painter = QtGui.QPainter()
        painter.begin(self.pix)
        rect = QtCore.QRect(0, 0, self.size + 4, self.size + 4)
        painter.eraseRect(rect)
        painter.end()
        self.setPixmap(self.pix)
        self.__set_pixmap()
        if self.shape == "square":
            self.__set_square()
        else:
            self.__set_circle()

    def __get_center(self):
        """Return the center of the scene.

        This assume both scene and image have the same geometry.

        Returns:
        list[center X, center Y, top left x, top left Y]
        """
        cX, cY = int(self.scene.width() /2), int(self.scene.height() /2)
        posX, posY = cX - int(self.size / 2), cY - int(self.size / 2)
        return [cX, cY, posX, posY]

    def __set_texture(self):
        """Return the sub-region of image at the current coordinate.

        Returns:
        QPixmap
        """
        texture = QtGui.QPixmap(QtCore.QSize(self.size + 1, self.size + 1))
        irect = QtCore.QRect(self.orig[2], self.orig[3], self.size + 1, self.size + 1)
        prect = QtCore.QRect(0, 0, self.size + 1, self.size + 1)
        painter = QtGui.QPainter()
        painter.begin(texture)
        painter.fillRect(prect, self.brush)
        painter.drawImage(prect, self.base, irect)
        painter.end()
        return texture

    def get_pos(self):
        """Return position of magnifying glass.

        Returns:
        QPointF(float, float)
        """
        return self.pos()

    def set_pos(self, pos):
        """Move magnifying glass to pos.

        Keyword arguments:
        pos -- QPointF
        """
        if isinstance(pos, QtCore.QPointF):
            X, Y = pos.x(), pos.y()
        else:
            X, Y = pos[0], pos[1]
        cur = self.pos()
        m_X, m_Y = X - self.pos().x(), Y - self.pos().y()
        self.moveBy(m_X, m_Y)
        self.__update(m_X, m_Y)
        self.positionChanged.emit([cur.x(), cur.y(), X, Y])

    def __get_orig_center(self):
        """Return the original position at the creation of the magnifier.

        Returns :
        list -- center of pixmap magnified and corner top left.
        """
        cX, cY = int(self.base.width() /2), int(self.base.height() /2)
        posX, posY = cX - int(self.size / 2), cY - int(self.size / 2)
        return [cX, cY, posX, posY]

    def recenter_magnifying_glass2(self):
        """Move the magnifying glass from center image to center graphicsView.

        """
        c = self.size / 2.0
        dest = self.graphic.mapToScene(self.graphic.width() / 2 - c,
                                        self.graphic.height() / 2 - c)
        deltaX = dest.x() - self.pos().x()
        deltaY = dest.y() - self.pos().y()
        j, k, l, m = deltaX / 50, deltaY / 50, self.pos().x(), self.pos().y()
        for i in range (50):
            l, m = l + j, m + k
            self.setPos(QtCore.QPointF(l, m))
            self.__update(j, k)
            QtCore.QCoreApplication.processEvents()

    def recenter_magnifying_glass(self):
        """Move the magnifying glass from center image to center graphicsView.

        """
        timer = QtCore.QTimeLine(1000);
        timer.setFrameRange(0, 50)
        self.animation = QtGui.QGraphicsItemAnimation()
        self.animation.setItem(self);
        self.animation.setTimeLine(timer)

        c = self.size / 2.0
        dest = self.graphic.mapToScene(self.graphic.width() / 2 - c, self.graphic.height() / 2 - c)
        deltaX = dest.x() - self.pos().x()
        deltaY = dest.y() - self.pos().y()
        j, k, l, m = deltaX / 50, deltaY / 50, self.pos().x(), self.pos().y()
        for i in range (50):
            l, m = l + j, m + k
            self.animation.setPosAt(i / 50.0, QtCore.QPointF(l, m))
            self.__update(j, k)
        timer.start()


    #--------------------------------
    # Events
    #--------------------------------

    def mousePressEvent(self, event, auto=0):
        # auto = 0 if event is caused by self
        # auto = 1 if event is given by other, i.e. event filter of main
        if auto:
            pos = self.graphic.mapToScene(event.pos())
        else:
            pos = event.scenePos()
        if event.button() == 1:
            if self.menu_wdg.isVisible():
                self.menu_wdg.hide()
            self.old_X, self.old_Y = pos.x(), pos.y()
            event.accept()
        else:
            event.ignore()


    def mouseMoveEvent(self, event, auto=0):
        # auto = 0 if event is caused by self
        # auto = 1 if event is given by other, i.e. event filter of main
        if auto:
            pos = self.graphic.mapToScene(event.pos())
        else:
            pos = event.scenePos()
        if event.button() == 2:
            return
        cur = self.pos()
        X, Y = pos.x(), pos.y()
        m_X, m_Y = X - self.old_X, Y - self.old_Y
        self.old_X, self.old_Y = X, Y
        self.moveBy(m_X, m_Y)
        self.__update(m_X, m_Y)
        event.accept()

    def contextMenuEvent(self, event):
        self.__show_settings(event.screenPos())
        event.accept()

    def hoverEnterEvent(self, event):
        shape = QtGui.QPixmap("medias/cross_cursor_32.png")
        cursor = QtGui.QCursor(shape, -1, -1)
        self.setCursor(cursor)

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_Escape:
            self.quit()

        elif event.key() == 16777234:
            # arrow LEFT
            self.move_by(-5, 0)
            self.positionChanged.emit([self.pos().x() + 5, self.pos().y(),
                                    self.pos().x(), self.pos().y()])

        elif event.key() == 16777235:
            # arrow UP
            self.move_by(0, -5)
            self.positionChanged.emit([self.pos().x(), self.pos().y() + 5,
                                    self.pos().x(), self.pos().y()])

        elif event.key() == 16777236:
            # arrow RIGHT
            self.move_by(5, 0)
            self.positionChanged.emit([self.pos().x() - 5, self.pos().y(),
                                    self.pos().x(), self.pos().y()])

        elif event.key() == 16777237:
            # arrow DOWN
            self.move_by(0, 5)
            self.positionChanged.emit([self.pos().x(), self.pos().y() - 5,
                                    self.pos().x(), self.pos().y()])

        elif event.key() == 83:
            # Key S used for enable / disable shadow effect
            try:
                self.main.shadow.setEnabled(not self.main.shadow.isEnabled())
            except:
                pass


    #------------------------------------
    # Widget used instead of context menu
    #------------------------------------

    def __show_settings(self, pos):
        """Show the context menu.

        """
        win_pos = self.main.window_geometry
        self.size_spb.setValue(self.size)
        self.magn_spb.setValue(self.factor * 100)
        self.square_rdo.blockSignals(True)
        self.circle_rdo.blockSignals(True)
        if self.shape == "square":
            self.square_rdo.setChecked(True)
        else:
            self.circle_rdo.setChecked(True)
        self.square_rdo.blockSignals(False)
        self.circle_rdo.blockSignals(False)
        self.menu_wdg.popup(pos)


    def build_settings(self):
        style = "".join(["QMenu {background:qlineargradient(",
                        "spread:pad, x1:1, y1:1, x2:1, y2:0, stop:0 ",
                        "rgba(100, 100, 100, 100), stop:1 rgba(255,",
                        "255, 255, 180)); ",
                        "border: 1px solid black; border-radius: 5px}"])
        self.menu_wdg = QtGui.QMenu()
        self.menu_wdg.resize(230, 140)
        self.menu_wdg.setMaximumSize(QtCore.QSize(285, 180))
        self.menu_wdg.setStyleSheet(style)
        self.gridLayout = QtGui.QGridLayout(self.menu_wdg)
        self.gridLayout.setSpacing(0)
        self.gridLayout.setContentsMargins(2, 0, 2, 0)
        self.verticalLayout = QtGui.QVBoxLayout()
        self.verticalLayout.setSpacing(0)
        font = QtGui.QFont()
        font.setFamily("Arial")
        font.setPointSize(10)
        font.setWeight(75)
        font.setBold(True)
        self.horizontalLayout = QtGui.QHBoxLayout()
        self.size_lbl = QtGui.QLabel(self.menu_wdg)
        self.size_lbl.setFont(font)
        self.horizontalLayout.addWidget(self.size_lbl)
        spacerItem = QtGui.QSpacerItem(288, 20, QtGui.QSizePolicy.Expanding,
                                    QtGui.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.horizontalLayout_3 = QtGui.QHBoxLayout()
        self.size_sld = QtGui.QSlider(self.menu_wdg)
        self.size_sld.setMinimumSize(QtCore.QSize(150, 0))
        self.size_sld.setMinimum(80)
        self.size_sld.setMaximum(400)
        self.size_sld.setSingleStep(10)
        self.size_sld.setProperty("value", 160)
        self.size_sld.setOrientation(QtCore.Qt.Horizontal)
        self.size_sld.setObjectName("size_sld")
        self.size_sld.setStyleSheet("QWidget {background-color: #CCCCCC}")
        self.horizontalLayout_3.addWidget(self.size_sld)
        self.size_spb = QtGui.QSpinBox(self.menu_wdg)
        self.size_spb.setMinimumSize(QtCore.QSize(70, 0))
        self.size_spb.setMaximumSize(QtCore.QSize(70, 20))
        self.size_spb.setSuffix(" px")
        self.size_spb.setMinimum(80)
        self.size_spb.setMaximum(400)
        self.size_spb.setSingleStep(10)
        self.size_spb.setProperty("value", 160)
        self.size_spb.setObjectName("size_spb")
        self.horizontalLayout_3.addWidget(self.size_spb)
        self.verticalLayout.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_4 = QtGui.QHBoxLayout()
        self.magn_lbl = QtGui.QLabel(self.menu_wdg)
        self.magn_lbl.setFont(font)
        self.horizontalLayout_4.addWidget(self.magn_lbl)
        spacerItem1 = QtGui.QSpacerItem(158, 20, QtGui.QSizePolicy.Expanding,
                                    QtGui.QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem1)
        self.verticalLayout.addLayout(self.horizontalLayout_4)
        self.horizontalLayout_5 = QtGui.QHBoxLayout()
        self.magn_sld = QtGui.QSlider(self.menu_wdg)
        self.magn_sld.setMinimumSize(QtCore.QSize(150, 0))
        self.magn_sld.setMinimum(50)
        self.magn_sld.setMaximum(600)
        self.magn_sld.setSingleStep(25)
        self.magn_sld.setProperty("value", 100)
        self.magn_sld.setOrientation(QtCore.Qt.Horizontal)
        self.magn_sld.setObjectName("magn_sld")
        self.magn_sld.setStyleSheet("QWidget {background-color: #A0A0A0}")
        self.horizontalLayout_5.addWidget(self.magn_sld)
        self.magn_spb = QtGui.QSpinBox(self.menu_wdg)
        self.magn_spb.setMinimumSize(QtCore.QSize(70, 0))
        self.magn_spb.setMaximumSize(QtCore.QSize(70, 20))
        self.magn_spb.setSuffix(" %")
        self.magn_spb.setMinimum(50)
        self.magn_spb.setMaximum(600)
        self.magn_spb.setSingleStep(20)
        self.magn_spb.setProperty("value", 100)
        self.magn_spb.setObjectName("magn_spb")
        self.horizontalLayout_5.addWidget(self.magn_spb)
        self.verticalLayout.addLayout(self.horizontalLayout_5)
        self.horizontalLayout_6 = QtGui.QHBoxLayout()
        self.shape_lbl = QtGui.QLabel(self.menu_wdg)
        self.shape_lbl.setFont(font)
        self.horizontalLayout_6.addWidget(self.shape_lbl)
        self.square_rdo = QtGui.QRadioButton(self.menu_wdg)
        self.square_rdo.setObjectName("square_rdo")
        self.square_rdo.setStyleSheet("QWidget {background-color: #707070}")
        self.horizontalLayout_6.addWidget(self.square_rdo)
        self.circle_rdo = QtGui.QRadioButton(self.menu_wdg)
        self.circle_rdo.setObjectName("circle_rdo")
        self.square_rdo.setStyleSheet("QWidget {background-color: #707070}")
        self.horizontalLayout_6.addWidget(self.circle_rdo)
        self.verticalLayout.addLayout(self.horizontalLayout_6)
        self.exit_btn = QtGui.QPushButton(self.menu_wdg)
        self.exit_btn.setObjectName("exit_btn")
        self.verticalLayout.addWidget(self.exit_btn)
        self.gridLayout.addLayout(self.verticalLayout, 0, 0, 1, 1)

        self.size_lbl.setText(QtGui.QApplication.translate("wdg",
                                    "Size in pixels", None,
                                    QtGui.QApplication.UnicodeUTF8))
        self.magn_lbl.setText(QtGui.QApplication.translate("wdg",
                                    "Magnification factor", None,
                                    QtGui.QApplication.UnicodeUTF8))
        self.shape_lbl.setText(QtGui.QApplication.translate("wdg",
                                    "Shape   ", None,
                                    QtGui.QApplication.UnicodeUTF8))
        self.square_rdo.setText(QtGui.QApplication.translate("wdg",
                                    "Square", None,
                                    QtGui.QApplication.UnicodeUTF8))
        self.circle_rdo.setText(QtGui.QApplication.translate("wdg",
                                    "Circle", None,
                                    QtGui.QApplication.UnicodeUTF8))
        self.exit_btn.setText(QtGui.QApplication.translate("wdg",
                                    "Close magnifying glass", None,
                                    QtGui.QApplication.UnicodeUTF8))

        self.size_sld.valueChanged.connect(self.size_slider_changed)
        self.size_spb.valueChanged.connect(self.size_spbox_changed)
        self.magn_sld.valueChanged.connect(self.magn_slider_changed)
        self.magn_spb.valueChanged.connect(self.magn_spbox_changed)
        self.circle_rdo.toggled.connect(self.send_settings)
        self.exit_btn.clicked.connect(self.quit)
        QtCore.QMetaObject.connectSlotsByName(self.menu_wdg)

    def size_slider_changed(self, val):
        """Set size value when size's slider has moved.

        keyword arguments:
        val -- size int()
        """
        self.size_spb.blockSignals(True)
        self.size_spb.setValue(val)
        self.size_spb.blockSignals(False)
        self.send_settings()

    def size_spbox_changed(self, val):
        """Set size value when size's spinBox has changed.

        keyword arguments:
        val -- size int()
        """
        self.size_sld.blockSignals(True)
        self.size_sld.setValue(val)
        self.size_sld.blockSignals(False)
        self.send_settings()

    def magn_slider_changed(self, val):
        """Set size value when factor's slider has moved.

        keyword arguments:
        val -- magnification factor int()
        """
        self.magn_spb.blockSignals(True)
        self.magn_spb.setValue(val)
        self.magn_spb.blockSignals(False)
        self.send_settings()

    def magn_spbox_changed(self, val):
        """Set size value when factor's spinbox has changed.

        keyword arguments:
        val -- magnification factor int()
        """
        self.magn_sld.blockSignals(True)
        self.magn_sld.setValue(val)
        self.magn_sld.blockSignals(False)
        self.send_settings()

    def send_settings(self):
        """Apply new settings.

        """
        if self.size != self.size_spb.value():
            self.set_size(self.size_spb.value())
        if self.factor != self.magn_spb.value() / 100.0:
            self.set_factor(self.magn_spb.value() / 100.0)
        if self.square_rdo.isChecked() and self.shape != 'square':
            self.set_shape("square")
        elif self.circle_rdo.isChecked() and self.shape != 'circle':
            self.set_shape("circle")
        self.settingsChanged.emit([self.factor, self.shape, self.size])

    def quit(self):
        """Quiet exit magnifying glass.

        """
        if self.menu_wdg.isVisible():
            self.menu_wdg.hide()
        self.scene.removeItem(self)


class Moved(QtCore.QObject):
    """Signal used by magnifier when magnifying glass position changed.

    """
    positionChanged = QtCore.pyqtSignal(list)
    def __init__(self):
        super(Moved, self).__init__()


class OptionChanged(QtCore.QObject):
    """Signal used by magnifier when magnifying glass settings are changed.

    """
    settingsChanged = QtCore.pyqtSignal(list)
    def __init__(self):
        super(OptionChanged, self).__init__()





